# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 https://aws.amazon.com/apache-2-0/


import boto3
import sys
import time
import os
import json

try:
    agent_config_file = os.environ["AGENT_CONFIG_FILE"]
except KeyError:
    agent_config_file = "/etc/agent/Agent_config.tfvars.json"

try:
    with open(agent_config_file, "r") as file:
        agent_config_data = json.loads(file.read())
except OSError:
    # This path is expected to be executed in Lambdas which don't have config
    # files
    agent_config_data = {
        "error_log_group": os.environ["ERROR_LOG_GROUP"],
        "error_logging_stream": os.environ["ERROR_LOGGING_STREAM"],
        "region": os.environ["REGION"],
    }


cw = boto3.client("logs", agent_config_data["region"])


# DEBUGGING
def log(
    message,
    log_group_name=agent_config_data["error_log_group"],
    log_stream_name=agent_config_data["error_logging_stream"],
):
    # print("ERROR-PRINT: {}".format(message))

    try:
        # retreive seq number
        response = cw.describe_log_streams(logGroupName=log_group_name)
        seq = None
        logEvents = [{"timestamp": int(time.time() * 1000), "message": message}]
        for lg in response.get("logStreams"):
            if lg.get("logStreamName") == log_stream_name:
                seq = lg.get("uploadSequenceToken")
        if seq is None:
            response = cw.put_log_events(
                logGroupName=log_group_name,
                logStreamName=log_stream_name,
                logEvents=logEvents,
            )
        else:
            response = cw.put_log_events(
                logGroupName=log_group_name,
                logStreamName=log_stream_name,
                logEvents=logEvents,
                sequenceToken=seq,
            )
    except Exception as e:
        print("Cannot log errors because {}".format(e), file=sys.stderr)



# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 https://aws.amazon.com/apache-2-0/


class MockComputeEngineJobWrapper:
    def __init__(self, worker_arguments, n_tasks_per_job, payload_options):
        self.worker_arguments = worker_arguments
        self.payload_options = payload_options
        self.n_tasks_per_job = n_tasks_per_job

        self.expected_output_string = ""

    def form_dict_task_definition(self):
        worker_arguments = self.worker_arguments.split(" ")

        task_definition = {"worker_arguments": worker_arguments}

        if int(worker_arguments[0]) >= 0:
            self.expected_output_string = worker_arguments[0]
        elif worker_arguments[-1] in self.computational_results_dic:
            self.expected_output_string = self.computational_results_dic[
                worker_arguments[-1]
            ]
        else:
            raise Exception(
                "Can not verify result as computation is not known for input '{}'".format(
                    self.worker_arguments
                )
            )

        return task_definition

    def generate_binary_job(self):
        task_dict = self.form_dict_task_definition()

        tasks = []

        for i in range(0, self.n_tasks_per_job):
            tasks.append(task_dict)

        return tasks

    def verify_results(self, stdout):
        if stdout.rstrip() != self.expected_output_string:
            return False, "Expected: [{}] != Received [{}]".format(
                self.expected_output_string, stdout.rstrip()
            )
        return True, ""



# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 https://aws.amazon.com/apache-2-0/


TASK_STATE_CANCELLED = "cancelled"
TASK_STATE_PENDING = "pending"
TASK_STATE_FAILED = "failed"
TASK_STATE_FINISHED = "finished"
TASK_STATE_PROCESSING = "processing"


class StateTableException(Exception):
    def __init__(
        self,
        original_message,
        supplied_message,
        caused_by_throttling=False,
        caused_by_condition=False,
    ):
        super().__init__(original_message)

        self.caused_by_throttling = caused_by_throttling
        self.caused_by_condition = caused_by_condition
        self.original_message = original_message
        self.supplied_message = supplied_message

    def __str__(self):
        return f"Original Message: {self.original_message}, supplied_message: {self.supplied_message}"



# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 https://aws.amazon.com/apache-2-0/


class TaskQueueException(Exception):
    def __init__(self, original_exception, supplied_message, traceback_msg):
        super().__init__(original_exception)
        self.original_message = str(original_exception)
        self.supplied_message = supplied_message
        self.traceback_msg = traceback_msg

    def get_original_exception(self):
        return self.get_original_exception

    def get_supplied_message(self):
        return self.supplied_message

    def get_traceback_message(self):
        return self.traceback_msg

    def __str__(self):
        return f"TaskQueueException: Original Message: {self.original_message}, supplied_message: {self.supplied_message}"



"""Real Python feed reader.

Import the `feed` module to work with the Real Python feed:

    >>> from reader import feed
    >>> feed.get_titles()
    ['Logging in Python', 'The Best Python Books', ...]

See https://github.com/realpython/reader/ for more information.
"""
# Standard library imports
# from importlib import resources

# try:
#     import tomllib
# except ModuleNotFoundError:
#     # Third party imports
#     import tomli as tomllib


# Version of kb-reader package
__version__ = "1.1.2"

# # Read URL of the Real Python feed from config file
# _cfg = tomllib.loads(resources.read_text("reader", "config.toml"))
# URL = _cfg["feed"]["url"]



# __main__.py

import sys

# from reader import feed, viewer

def main():
    """Read the Real Python article feed"""
    print ("Main Package")

    # # If an article ID is given, then show the article
    # if len(sys.argv) > 1:
    #     article = feed.get_article(sys.argv[1])
    #     viewer.show(article)

    # # If no ID is given, then show a list of all articles
    # else:
    #     site = feed.get_site()
    #     titles = feed.get_titles()
    #     viewer.show_list(site, titles)

if __name__ == "__main__":
    main()


# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 https://aws.amazon.com/apache-2-0/

import json
import datetime
import random
import boto3


class PerfTrackerFirehoseConnector:
    def __init__(self, connector_string):
        """
        Expected format of the connection string:
        "<aws_region_name> <firehose_stream_name>"
        example:
        "eu-west-1 fh_submit_tasks_lambda"
        """

        tokens = connector_string.split(" ")
        self.region_name = tokens[0]
        self.delivery_stream_name = tokens[1]

        # self.firehose_client = boto3.client('firehose',  region_name=self.region_name)
        self.firehose_client = boto3.client(
            "firehose",
            region_name=self.region_name,
            endpoint_url=f"https://firehose.{self.region_name}.amazonaws.com",
        )

        self.samples_buffer = []

    def __build_msg(self):
        return {
            "measurement": "event1",
            "time": datetime.datetime.utcnow(),
            "fields": {
                "duration": random.randint(0, 1000),
            },
        }

    def add_sample(self, json_data_sample):
        sample = {"Data": json.dumps(json_data_sample)}

        self.samples_buffer.append(sample)

    def submit_measurements(self):
        res = self.firehose_client.put_record_batch(
            DeliveryStreamName=self.delivery_stream_name, Records=self.samples_buffer
        )

        self.samples_buffer = []

        return res



# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 https://aws.amazon.com/apache-2-0/

import random
import time


class TTLExpirationGenerator:
    def __init__(self, task_ttl_refresh_interval_sec, task_ttl_expiration_offset_sec):
        # This is how often we send heart beats, i.e., how often we extend TTL value
        self.refresh_interval_sec = task_ttl_refresh_interval_sec
        # This is by how much inot the future we extend TTL value
        self.expriation_offset_sec = task_ttl_expiration_offset_sec

        if self.refresh_interval_sec >= self.expriation_offset_sec:
            raise Exception(
                """Refresh interval [{}] must be smaller then expiration offset [{}].
                Otherwise TTL will always expire before we have chance to extend it!""".format(
                    self.refresh_interval_sec, self.expriation_offset_sec
                )
            )
        self.next_refresh_timestamp = 0
        self.next_expiration_timestamp = 0

    def generate_next_ttl(self):
        self.next_refresh_timestamp = int(time.time()) + random.randint(
            self.refresh_interval_sec, int(self.refresh_interval_sec * 1.1)
        )
        self.next_expiration_timestamp = self.next_refresh_timestamp + (
            self.expriation_offset_sec - self.refresh_interval_sec
        )
        return self

    def get_next_expiration_timestamp(self):
        return self.next_expiration_timestamp

    def get_next_refresh_timestamp(self):
        return self.next_refresh_timestamp



# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 https://aws.amazon.com/apache-2-0/

from influxdb import InfluxDBClient


class PerfTrackerInfluxDBConnector:
    def __init__(self, connector_string, influxdb_ip):
        """
        Expected format of the connection string:
        "<aws_region_name> <firehose_stream_name>"
        example:
        "eu-west-1 fh_submit_tasks_lambda"
        """

        tokens = connector_string.split(" ")
        self.influxdb_ip = influxdb_ip
        self.influxdb_port = tokens[0]
        self.database = tokens[1]
        self.measurement = tokens[2]

        print(self.influxdb_ip, self.influxdb_port, self.database, self.measurement)

        # self.firehose_client = boto3.client('firehose',  region_name=self.region_name)
        # self.firehose_client = boto3.client('firehose',  region_name=self.region_name, endpoint_url="https://firehose.eu-west-1.amazonaws.com")

        self.influxdb_client = InfluxDBClient(
            host=self.influxdb_ip, port=self.influxdb_port
        )

        self.influxdb_client.create_database(self.database)
        self.influxdb_client.switch_database(self.database)

        self.samples_buffer = []

    def add_sample(self, json_data_sample):
        fields = {}

        for k, v in json_data_sample.items():
            fields[k] = v

        sample = {
            "measurement": self.measurement,
            "time": json_data_sample["EVENT_TIME"],
            "fields": fields,
        }

        self.samples_buffer.append(sample)

    def submit_measurements(self):
        print(self.samples_buffer)
        res = self.influxdb_client.write_points(self.samples_buffer)

        self.samples_buffer = []

        return res



# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 https://aws.amazon.com/apache-2-0/

import datetime
import time

from htcutils.perf_tracker_firehose_connector import PerfTrackerFirehoseConnector
from htcutils.perf_tracker_influxdb_connector import PerfTrackerInfluxDBConnector


def get_time_now_ms():
    return int(round(time.time() * 1000))


# TODO: remove dependencies on influxdb
def performance_tracker_initializer(
    metrics_are_enabled, connection_string, influxdb_ip
):
    metrics_are_enabled = bool(int(metrics_are_enabled))
    if metrics_are_enabled:
        tokens = connection_string.split(" ", 1)  # Pick up first word in the string
        connector_type = tokens[0]
        if connector_type == "firehose":
            firehost_connector = PerfTrackerFirehoseConnector(
                connector_string=tokens[1]
            )
            perf_tracker = PerformanceTracker(firehost_connector)
            return perf_tracker
        if connector_type == "influxdb":
            influxdb_connector = PerfTrackerInfluxDBConnector(
                connector_string=tokens[1], influxdb_ip=influxdb_ip
            )
            perf_tracker = PerformanceTracker(influxdb_connector)
            return perf_tracker
        else:
            print(
                "ERROR Undefined metrics connector type, no metrics will be collected: {}".format(
                    connector_type
                )
            )
            return __create_empty_performance_tracker
    else:
        return __create_empty_performance_tracker()


def __create_empty_performance_tracker():
    return PerformanceTracker(None)


class EventsCounter:
    def __init__(self, expected_events=None):
        if expected_events is None:
            expected_events = []
        self.expected_events = expected_events
        self.evcounter = {}
        self.reset()

    def increment(self, event_name, value=1):
        if event_name in self.evcounter:
            self.evcounter[event_name] += value
        else:
            self.evcounter[event_name] = 1

    def set(self, event_name, value):
        self.evcounter[event_name] = value

    def get_counter(self, name):
        if name in self.evcounter:
            return self.evcounter[name]
        elif name.startswith("str"):
            return ""
        else:
            return 0

    def reset(self):
        """When resetting counter we need to prebuild keys for expected counters,
        because elastic search can not change index mapping when new counters are added"""

        self.evcounter = {}

        for e in self.expected_events:
            if e.startswith("str"):
                self.evcounter[e] = ""
            else:
                self.evcounter[e] = 0


class PerformanceTracker:
    def __init__(self, buffered_storage_connector):
        self.buffered_storage_connector = buffered_storage_connector

        self.stats_batch = []

        self.last_batch_submission_delay_ms = 0
        self.last_batch_submission_timestamp_ms = 0
        self.max_batching_delay_ms = 5 * 1000

    def add_metric_sample(
        self, stats_dic, event_counter, from_event, to_event, event_time=None
    ):
        if not event_time:
            event_time = datetime.datetime.now().isoformat()

        # print(stats_dic)
        sorted_keys = sorted(stats_dic.keys())
        # for k in sorted_keys:
        #     print(k)
        start_index = sorted_keys.index(from_event)
        end_index = sorted_keys.index(to_event)

        data = {"EVENT_TIME": event_time}

        for i in range(start_index, end_index):
            # print([sorted_keys[i]])
            # print(stats_dic[sorted_keys[i]])
            # print([sorted_keys[i+1]])
            # print(stats_dic[sorted_keys[i+1]])
            # print("--------")
            key = stats_dic[sorted_keys[i + 1]]["label"]
            value = (
                stats_dic[sorted_keys[i + 1]]["tstmp"]
                - stats_dic[sorted_keys[i]]["tstmp"]
            )
            data[key] = value

        if event_counter is not None:
            for key, value in sorted(event_counter.evcounter.items()):
                data[key] = value

            event_counter.reset()

        # Self profiling on the batch submission delays
        data["last_batch_submission_delay_ms"] = self.last_batch_submission_delay_ms

        if self.buffered_storage_connector:
            self.buffered_storage_connector.add_sample(data)

    def submit_measurements(self):
        if self.buffered_storage_connector:
            if (
                self.max_batching_delay_ms
                < get_time_now_ms() - self.last_batch_submission_timestamp_ms
            ):
                time_start_ms = get_time_now_ms()
                self.buffered_storage_connector.submit_measurements()
                self.last_batch_submission_delay_ms = get_time_now_ms() - time_start_ms
                self.last_batch_submission_timestamp_ms = get_time_now_ms()
